"""Tests of scriv/ghrel.py."""

import json
import logging
from textwrap import dedent
from typing import Any
from unittest.mock import call

import pytest

from .helpers import check_logs

CHANGELOG1 = dedent("""\
    Some text before

    v1.2.3 -- 2022-04-21
    --------------------

    A good release

    Some fixes
    ----------

    No version number in this section.

    v1.0 -- 2020-02-20
    ------------------

    Nothing to say.

    v0.9a7 -- 2017-06-16
    --------------------

    A beginning

    v0.1 -- 2010-01-01
    ------------------

    Didn't bother to tag this one.

    v0.0.1 -- 2001-01-01
    --------------------

    Very first.

    """)

RELEASES1 = {
    "v1.0": {
        "url": "https://api.github.com/repos/joe/project/releases/120",
        "body": "Nothing to say.\n",
    },
    "v0.9a7": {
        "url": "https://api.github.com/repos/joe/project/releases/123",
        "body": "original body",
    },
    "v0.0.1": {
        "url": "https://api.github.com/repos/joe/project/releases/123",
        "body": "original body",
    },
}

V123_REL = {
    "body": "A good release\n",
    "name": "v1.2.3",
    "tag_name": "v1.2.3",
    "draft": False,
    "prerelease": False,
}

V097_REL = {
    "body": "A beginning\n",
    "name": "v0.9a7",
    "tag_name": "v0.9a7",
    "draft": False,
    "prerelease": True,
}

V001_REL = {
    "body": "Very first.\n",
    "name": "v0.0.1",
    "tag_name": "v0.0.1",
    "draft": False,
    "prerelease": False,
}


@pytest.fixture()
def scenario1(temp_dir, fake_git, mocker):
    """A common scenario for the tests."""
    fake_git.add_remote("origin", "git@github.com:joe/project.git")
    fake_git.add_tags(["v1.2.3", "v1.0", "v0.9a7", "v0.0.1"])
    (temp_dir / "CHANGELOG.rst").write_text(CHANGELOG1)
    mock_get_releases = mocker.patch("scriv.ghrel.get_releases")
    mock_get_releases.return_value = RELEASES1


@pytest.fixture()
def mock_create_release(mocker):
    """Create a mock create_release that checks arguments."""

    def _create_release(repo: str, release_data: dict[str, Any]) -> None:
        assert repo
        assert release_data["name"]
        assert json.dumps(release_data)[0] == "{"

    return mocker.patch(
        "scriv.ghrel.create_release", side_effect=_create_release
    )


@pytest.fixture()
def mock_update_release(mocker):
    """Create a mock update_release that checks arguments."""

    def _update_release(
        release: dict[str, Any], release_data: dict[str, Any]
    ) -> None:
        assert release_data["name"]
        assert release["url"]
        assert json.dumps(release_data)[0] == "{"

    return mocker.patch(
        "scriv.ghrel.update_release", side_effect=_update_release
    )


def test_default(
    cli_invoke, scenario1, mock_create_release, mock_update_release, caplog
):
    cli_invoke(["github-release"])

    assert mock_create_release.mock_calls == [call("joe/project", V123_REL)]
    assert mock_update_release.mock_calls == []
    assert caplog.record_tuples == [
        (
            "scriv.changelog",
            logging.INFO,
            "Reading changelog CHANGELOG.rst",
        ),
    ]


def test_default_draft(
    cli_invoke, scenario1, mock_create_release, mock_update_release, caplog
):
    cli_invoke(["github-release", "--draft"])

    draft_rel = V123_REL.copy()
    draft_rel["draft"] = True

    assert mock_create_release.mock_calls == [call("joe/project", draft_rel)]
    assert mock_update_release.mock_calls == []
    assert caplog.record_tuples == [
        (
            "scriv.changelog",
            logging.INFO,
            "Reading changelog CHANGELOG.rst",
        ),
    ]


def test_dash_all(
    cli_invoke, scenario1, mock_create_release, mock_update_release, caplog
):
    cli_invoke(["github-release", "--all"])

    assert mock_create_release.mock_calls == [call("joe/project", V123_REL)]
    assert mock_update_release.mock_calls == [
        call(RELEASES1["v0.9a7"], V097_REL),
        call(RELEASES1["v0.0.1"], V001_REL),
    ]
    assert caplog.record_tuples == [
        (
            "scriv.changelog",
            logging.INFO,
            "Reading changelog CHANGELOG.rst",
        ),
        (
            "scriv.ghrel",
            logging.WARNING,
            "Entry 'Some fixes' has no version, skipping.",
        ),
        (
            "scriv.ghrel",
            logging.WARNING,
            "Version v0.1 has no tag. No release will be made.",
        ),
    ]


def test_explicit_repo(
    cli_invoke, scenario1, fake_git, mock_create_release, mock_update_release
):
    # Add another GitHub remote, now there are two.
    fake_git.add_remote("upstream", "git@github.com:psf/project.git")

    cli_invoke(["github-release", "--repo=xyzzy/plugh"])

    assert mock_create_release.mock_calls == [call("xyzzy/plugh", V123_REL)]
    assert mock_update_release.mock_calls == []


@pytest.mark.parametrize(
    "repo", ["xyzzy", "https://github.com/xyzzy/plugh.git"]
)
def test_bad_explicit_repo(cli_invoke, repo):
    result = cli_invoke(["github-release", f"--repo={repo}"], expect_ok=False)
    assert result.exit_code == 1
    assert str(result.exception) == f"Repo must be owner/reponame: {repo!r}"


def test_check_links(cli_invoke, scenario1, mocker):
    mock_check_links = mocker.patch("scriv.ghrel.check_markdown_links")
    cli_invoke(["github-release", "--all", "--dry-run", "--check-links"])
    assert mock_check_links.mock_calls == [
        call("A good release\n"),
        call("Nothing to say.\n"),
        call("A beginning\n"),
        call("Very first.\n"),
    ]


@pytest.fixture()
def no_actions(mock_create_release, mock_update_release, responses):
    """Check that nothing really happened."""

    yield

    assert mock_create_release.mock_calls == []
    assert mock_update_release.mock_calls == []
    assert len(responses.calls) == 0


def test_default_dry_run(cli_invoke, scenario1, no_actions, caplog):
    cli_invoke(["github-release", "--dry-run"])
    check_logs(
        caplog,
        [
            (
                "scriv.changelog",
                logging.INFO,
                "Reading changelog CHANGELOG.rst",
            ),
            ("scriv.ghrel", logging.INFO, "Would create release v1.2.3"),
        ],
    )


def test_dash_all_dry_run(cli_invoke, scenario1, no_actions, caplog):
    cli_invoke(["github-release", "--all", "--dry-run"])
    check_logs(
        caplog,
        [
            (
                "scriv.changelog",
                logging.INFO,
                "Reading changelog CHANGELOG.rst",
            ),
            ("scriv.ghrel", logging.INFO, "Would create release v1.2.3"),
            (
                "scriv.ghrel",
                logging.WARNING,
                "Entry 'Some fixes' has no version, skipping.",
            ),
            ("scriv.ghrel", logging.INFO, "Would update release v0.9a7"),
            (
                "scriv.ghrel",
                logging.WARNING,
                "Version v0.1 has no tag. No release will be made.",
            ),
            ("scriv.ghrel", 20, "Would update release v0.0.1"),
        ],
    )


def test_dash_all_dry_run_debug(cli_invoke, scenario1, no_actions, caplog):
    cli_invoke(["github-release", "--all", "--dry-run", "--verbosity=debug"])
    check_logs(
        caplog,
        [
            (
                "scriv.changelog",
                logging.INFO,
                "Reading changelog CHANGELOG.rst",
            ),
            (
                "scriv.ghrel",
                logging.DEBUG,
                "Creating release, data = {'body': 'A good release\\n', 'name': 'v1.2.3', "
                + "'tag_name': 'v1.2.3', 'draft': False, 'prerelease': False}",
            ),
            ("scriv.ghrel", logging.INFO, "Would create release v1.2.3"),
            ("scriv.ghrel", logging.DEBUG, "Body:\nA good release\n"),
            (
                "scriv.ghrel",
                logging.WARNING,
                "Entry 'Some fixes' has no version, skipping.",
            ),
            (
                "scriv.ghrel",
                logging.DEBUG,
                "Updating release v0.9a7, data = {'body': 'A beginning\\n', 'name': "
                + "'v0.9a7', 'tag_name': 'v0.9a7', 'draft': False, 'prerelease': True}",
            ),
            ("scriv.ghrel", logging.INFO, "Would update release v0.9a7"),
            ("scriv.ghrel", logging.DEBUG, "Old body:\noriginal body"),
            ("scriv.ghrel", logging.DEBUG, "New body:\nA beginning\n"),
            (
                "scriv.ghrel",
                logging.WARNING,
                "Version v0.1 has no tag. No release will be made.",
            ),
            (
                "scriv.ghrel",
                logging.DEBUG,
                "Updating release v0.0.1, data = {'body': 'Very first.\\n', 'name': "
                + "'v0.0.1', 'tag_name': 'v0.0.1', 'draft': False, 'prerelease': False}",
            ),
            ("scriv.ghrel", logging.INFO, "Would update release v0.0.1"),
            ("scriv.ghrel", logging.DEBUG, "Old body:\noriginal body"),
            ("scriv.ghrel", logging.DEBUG, "New body:\nVery first.\n"),
        ],
    )


def test_no_github_repo(cli_invoke, scenario1, fake_git):
    fake_git.remove_remote("origin")
    result = cli_invoke(["github-release"], expect_ok=False)
    assert result.exit_code == 1
    assert result.output == "Couldn't find a GitHub repo\n"


def test_no_clear_github_repo(cli_invoke, scenario1, fake_git):
    # Add another GitHub remote, now there are two.
    fake_git.add_remote("upstream", "git@github.com:psf/project.git")
    result = cli_invoke(["github-release"], expect_ok=False)
    assert result.exit_code == 1
    assert result.output == (
        "More than one GitHub repo found: joe/project, psf/project\n"
    )


def test_with_template(cli_invoke, temp_dir, scenario1, mock_create_release):
    (temp_dir / "setup.cfg").write_text(
        """
        [scriv]
        ghrel_template = |{{title}}|{{body}}|{{config.format}}|{{version}}|
        """
    )
    cli_invoke(["github-release"])

    expected = dict(V123_REL)
    expected["body"] = "|v1.2.3 -- 2022-04-21|A good release\n|rst|v1.2.3|"

    assert mock_create_release.mock_calls == [call("joe/project", expected)]



"""Helpers for using subprocesses."""

import logging
import shlex
import subprocess

# The return value of run_command.
CmdResult = tuple[bool, str]

logger = logging.getLogger(__name__)


def run_command(cmd: str | list[str]) -> CmdResult:
    """
    Run a command line (with no shell).

    Returns a tuple:
        bool: true if the command succeeded.
        str: the output of the command.

    """
    logger.debug(f"Running command {cmd!r}")
    if isinstance(cmd, str):
        cmd = shlex.split(cmd, posix=False)
    proc = subprocess.run(
        cmd,
        shell=False,
        check=False,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    output = proc.stdout.decode("utf-8")
    logger.debug(
        f"Command exited with {proc.returncode} status. Output: {output!r}"
    )

    return proc.returncode == 0, output


def run_simple_command(cmd: str | list[str]) -> str:
    """
    Run a command and return its output, or "" if it fails.
    """
    ok, out = run_command(cmd)
    if not ok:
        return ""
    return out.strip()


def run_shell_command(cmd: str) -> CmdResult:
    """
    Run a command line with a shell.
    """
    logger.debug(f"Running shell command {cmd!r}")
    proc = subprocess.run(
        cmd,
        shell=True,
        check=False,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    output = proc.stdout.decode("utf-8")
    logger.debug(
        f"Command exited with {proc.returncode} status. Output: {output!r}"
    )
    return proc.returncode == 0, output



"""Get information from git."""

import logging
import os
import re
import subprocess
import sys
from pathlib import Path

import click

from .shell import run_simple_command

logger = logging.getLogger(__name__)


def user_nick() -> str:
    """
    Get a short name for the current user.
    """
    nick = git_config("scriv.user-nick")
    if nick:
        return nick

    nick = git_config("github.user")
    if nick:
        return nick

    email = git_config("user.email")
    if email:
        nick = email.partition("@")[0]
        return nick

    return os.getenv("USER", "somebody")


def current_branch_name() -> str:
    """
    Get the current branch name.
    """
    return run_simple_command("git rev-parse --abbrev-ref HEAD")


def git_config(option: str) -> str:
    """
    Return a git config value.
    """
    return run_simple_command(f"git config --get {option}")


def git_config_bool(option: str) -> bool:
    """
    Return a boolean git config value, defaulting to False.
    """
    return git_config(option) == "true"


def git_edit(filename: Path) -> None:
    """Edit a file using the same editor Git chooses."""
    git_editor = run_simple_command("git var GIT_EDITOR")
    click.edit(filename=str(filename), editor=git_editor)


def git_add(filename: Path) -> None:
    """Git add a file. If it fails, sys.exit."""
    ret = subprocess.call(["git", "add", str(filename)])
    if ret == 0:
        logger.info(f"Added {filename}")
    else:
        logger.error(f"Couldn't add {filename}")
        sys.exit(ret)


def git_rm(filename: Path) -> None:
    """Git rm a file. If it fails, sys.exit."""
    ret = subprocess.call(["git", "rm", str(filename)])
    if ret == 0:
        logger.info(f"Removed {filename}")
    else:
        logger.error(f"Couldn't remove {filename}")
        sys.exit(ret)


def get_github_repos() -> set[str]:
    """
    Find the GitHub name/repos for this project.

    Returns a set of "name/repo" addresses for GitHub repos.
    """
    urls = run_simple_command("git remote -v").splitlines()
    github_repos = set()
    for url in urls:
        m = re.search(r"github.com[:/]([^/]+/\S+)", url)
        if m:
            repo = m[1]
            # It might or might not have .git appended.
            if repo.endswith(".git"):
                repo = repo[:-4]
            github_repos.add(repo)
    return github_repos



"""
Third-party modules that might or might not be available.
"""

# pylint: disable=unused-import

from types import ModuleType

tomllib: ModuleType | None

try:
    try:
        import tomllib  # type: ignore[no-redef]
    except ModuleNotFoundError:
        import tomli as tomllib  # type: ignore[no-redef]
except ImportError:
    tomllib = None


yaml: ModuleType | None

try:
    import yaml  # type: ignore[no-redef]
except ImportError:
    yaml = None



"""Creating fragments."""

import logging
import sys

import click

from .gitinfo import git_add, git_config_bool, git_edit
from .scriv import Scriv
from .util import scriv_command

logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "--add/--no-add", default=None, help="'git add' the created file."
)
@click.option(
    "--edit/--no-edit",
    default=None,
    help="Open the created file in your text editor.",
)
@scriv_command
def create(
    *,
    add: bool | None,
    edit: bool | None,
    config_file: str | None,
) -> None:
    """
    Create a new changelog fragment.
    """
    if add is None:
        add = git_config_bool("scriv.create.add")
    if edit is None:
        edit = git_config_bool("scriv.create.edit")
    scriv = Scriv(config_file=config_file)
    frag = scriv.new_fragment()
    file_path = frag.path
    if not file_path.parent.exists():
        sys.exit(
            f"Output directory {str(file_path.parent)!r} doesn't exist,"
            + " please create it."
        )

    if file_path.exists():
        sys.exit(f"File {file_path} already exists, not overwriting")

    logger.info(f"Creating {file_path}")
    frag.write()

    if edit:
        git_edit(file_path)
        sections = scriv.sections_from_fragment(frag)
        if not sections:
            logger.info("Empty fragment, aborting...")
            file_path.unlink()
            sys.exit()

    if add:
        git_add(file_path)



"""Helpers for the GitHub REST API."""

import logging
import os
from collections.abc import Iterable
from typing import Any

import requests

logger = logging.getLogger(__name__)


# Only wait up to a minute for GitHub to respond.
TIMEOUT = 60


def check_ok(resp):
    """
    Check that the Requests response object was successful.

    Raise an exception if not.
    """
    if not resp:
        logger.error(f"text: {resp.text!r}")
        resp.raise_for_status()


def auth_headers() -> dict[str, str]:
    """
    Get the authorization headers needed for GitHub.

    Will read the GITHUB_TOKEN environment variable.
    """
    headers = {}
    token = os.environ.get("GITHUB_TOKEN", "")
    if token:
        headers["Authorization"] = f"Bearer {token}"
    return headers


def github_paginated(url: str) -> Iterable[dict[str, Any]]:
    """
    Get all the results from a paginated GitHub url.
    """
    while True:
        resp = requests.get(url, headers=auth_headers(), timeout=TIMEOUT)
        check_ok(resp)
        yield from resp.json()
        next_link = resp.links.get("next", None)
        if not next_link:
            break
        url = next_link["url"]


RELEASES_URL = "https://api.github.com/repos/{repo}/releases"


def get_releases(repo: str) -> dict[str, dict[str, Any]]:
    """
    Get all the releases from a name/project repo.

    Returns:
        A dict mapping tag names to release dictionaries.
    """
    url = RELEASES_URL.format(repo=repo)
    releases = {r["tag_name"]: r for r in github_paginated(url)}
    return releases


def create_release(repo: str, release_data: dict[str, Any]) -> None:
    """
    Create a GitHub release.

    Arguments:
        repo: A user/repo string, like "nedbat/scriv".
        release_data: A dict with the data needed to create the release.
            It should have these keys:
                body: the markdown description of the release.
                name: the name of the release
                tag_name: the Git tag for the release
                draft: a boolean
                prerelease: a boolean
    """
    logger.info(f"Creating release {release_data['name']}")
    url = RELEASES_URL.format(repo=repo)
    resp = requests.post(
        url, json=release_data, headers=auth_headers(), timeout=TIMEOUT
    )
    check_ok(resp)


def update_release(
    release: dict[str, Any], release_data: dict[str, Any]
) -> None:
    """
    Update a GitHub release.

    Arguments:
        release: the full data from GitHub for the release to update.
        release_data: a dict with the data we want to update.
            See create_release for the accepted keys.
    """
    logger.info(f"Updating release {release_data['name']}")
    resp = requests.patch(
        release["url"],
        json=release_data,
        headers=auth_headers(),
        timeout=TIMEOUT,
    )
    check_ok(resp)



"""Tests of scriv/linkcheck.py"""

import logging
import textwrap

import pytest

from scriv.linkcheck import check_markdown_links, find_links


@pytest.mark.parametrize(
    "markdown_text, links",
    [
        ("Hello", []),
        (
            """\
            [one](https://two.com/hello) and
            [two](https://one.com/xyzzy).
         """,
            ["https://one.com/xyzzy", "https://two.com/hello"],
        ),
        (
            """\
            This is [an example](http://example1.com/ "Title") inline link.
            This is [an example] [id] reference-style link.

            [id]: http://example2.com/  "Optional Title Here"
         """,
            ["http://example1.com/", "http://example2.com/"],
        ),
    ],
)
def test_find_links(markdown_text, links):
    found_links = sorted(find_links(textwrap.dedent(markdown_text)))
    assert links == found_links


def test_check_markdown_link(caplog, responses):
    caplog.set_level(logging.DEBUG, logger="scriv.linkcheck")
    responses.head("https://nedbat.com")
    check_markdown_links("""[hey](https://nedbat.com)!""")
    assert caplog.record_tuples == [
        (
            "scriv.linkcheck",
            logging.DEBUG,
            "OK link: 'https://nedbat.com'",
        )
    ]


def test_check_404_markdown_link(caplog, responses):
    responses.head("https://nedbat.com", status=404)
    check_markdown_links("""[hey](https://nedbat.com)!""")
    assert caplog.record_tuples == [
        (
            "scriv.linkcheck",
            logging.WARNING,
            "Failed check for 'https://nedbat.com': status code 404",
        )
    ]


def test_check_failing_markdown_link(caplog, responses):
    responses.head("https://nedbat.com", body=Exception("Buh?"))
    check_markdown_links("""[hey](https://nedbat.com)!""")
    assert caplog.record_tuples == [
        (
            "scriv.linkcheck",
            logging.WARNING,
            "Failed check for 'https://nedbat.com': Buh?",
        )
    ]



"""Miscellanous helpers."""

from __future__ import annotations

import collections
import functools
import logging
import re
import sys
from collections.abc import Sequence
from typing import TypeVar

import click
import click_log

from .exceptions import ScrivException

T = TypeVar("T")
K = TypeVar("K")


def order_dict(
    d: dict[K | None, T], keys: Sequence[K | None]
) -> dict[K | None, T]:
    """
    Produce an OrderedDict of `d`, but with the keys in `keys` order.

    Keys in `d` that don't appear in `keys` will be at the end in an
    undetermined order.
    """
    with_order = collections.OrderedDict()
    to_insert = set(d)
    for k in keys:
        if k not in to_insert:
            continue
        with_order[k] = d[k]
        to_insert.remove(k)

    for k in to_insert:
        with_order[k] = d[k]

    return with_order


def partition_lines(text: str, marker: str) -> tuple[str, str, str]:
    """
    Split `text` by lines, similar to str.partition.

    The splitting line is the first line containing `marker`.

    """
    lines = text.splitlines(keepends=True)
    marker_pos = [i for i, line in enumerate(lines) if marker in line]
    if not marker_pos:
        return (text, "", "")
    pos = marker_pos[0]
    return (
        "".join(lines[:pos]),
        lines[pos],
        "".join(lines[pos + 1 :]),
    )


VERSION_REGEX = r"""(?ix)   # based on https://peps.python.org/pep-0440/
    \b                      # at a word boundary
    v?                      # maybe a leading "v"
    (\d+!)?                 # maybe a version epoch
    \d+(\.\d+)+             # the meat of the version number: N.N.N
    (?P<pre>
        [-._]?[a-z]+\.?\d*
    )?                      # maybe a pre-release: .beta3
    ([-._][a-z]+\d*)*       # maybe post and dev releases
    (\+\w[\w.]*\w)?         # maybe a local version
    \b
    """


class Version:
    """
    A version string that compares correctly.

    For example, "v1.2.3" and "1.2.3" are considered the same.

    """

    def __init__(self, vtext: str) -> None:
        """Create a smart version from a string version number."""
        self.vtext = vtext

    def __repr__(self):
        return f"<Version {self.vtext!r}>"

    def __str__(self):
        return self.vtext

    def __bool__(self):
        return bool(self.vtext)

    def __eq__(self, other):
        assert isinstance(other, Version)
        return self.vtext.lstrip("v") == other.vtext.lstrip("v")

    def __hash__(self):
        return hash(self.vtext.lstrip("v"))

    @classmethod
    def from_text(cls, text: str) -> Version | None:
        """Find a version number in a text string."""
        m = re.search(VERSION_REGEX, text)
        if m:
            return cls(m[0])
        return None

    def is_prerelease(self) -> bool:  # noqa: D400
        """Is this version number a pre-release?"""
        m = re.fullmatch(VERSION_REGEX, self.vtext)
        assert m  # the version must be a valid version
        return bool(m["pre"])


def scriv_command(func):
    """
    Decorate scriv commands to provide scriv-specific behavior.

    - ScrivExceptions don't show tracebacks

    - Set the log level from the command line for all of scriv.

    - All commands take a --config option.

    """

    @functools.wraps(func)
    @click_log.simple_verbosity_option(logging.getLogger("scriv"))
    @click.option(
        "--config",
        "config_file",
        default=None,
        type=click.Path(exists=True, readable=True, dir_okay=False),
        help="Use a custom config file.",
    )
    def _wrapped(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except ScrivException as exc:
            sys.exit(str(exc))

    return _wrapped



"""Tests of gitinfo.py"""

import re

from scriv.gitinfo import current_branch_name, get_github_repos, user_nick


def test_user_nick_from_scriv_user_nick(fake_git):
    fake_git.set_config("scriv.user-nick", "joedev")
    assert user_nick() == "joedev"


def test_user_nick_from_github(fake_git):
    fake_git.set_config("github.user", "joedev")
    assert user_nick() == "joedev"


def test_user_nick_from_git(fake_git):
    fake_git.set_config("user.email", "joesomeone@somewhere.org")
    assert user_nick() == "joesomeone"


def test_user_nick_from_env(fake_git, monkeypatch):
    monkeypatch.setenv("USER", "joseph")
    assert user_nick() == "joseph"


def test_user_nick_from_nowhere(fake_git, monkeypatch):
    # With no git information, and no USER env var,
    # we just call the user "somebody"
    monkeypatch.delenv("USER", raising=False)
    assert user_nick() == "somebody"


def test_current_branch_name(fake_git):
    fake_git.set_branch("joedev/feature-123")
    assert current_branch_name() == "joedev/feature-123"


def test_get_github_repos_no_remotes(fake_git):
    assert get_github_repos() == set()


def test_get_github_repos_one_github_remote(fake_git):
    fake_git.add_remote("mygithub", "git@github.com:joe/myproject.git")
    assert get_github_repos() == {"joe/myproject"}


def test_get_github_repos_one_github_remote_no_extension(fake_git):
    fake_git.add_remote("mygithub", "git@github.com:joe/myproject")
    assert get_github_repos() == {"joe/myproject"}


def test_get_github_repos_two_github_remotes(fake_git):
    fake_git.add_remote("mygithub", "git@github.com:joe/myproject.git")
    fake_git.add_remote("upstream", "git@github.com:psf/myproject.git")
    assert get_github_repos() == {"joe/myproject", "psf/myproject"}


def test_get_github_repos_one_github_plus_others(fake_git):
    fake_git.add_remote("mygithub", "git@github.com:joe/myproject.git")
    fake_git.add_remote("upstream", "git@gitlab.com:psf/myproject.git")
    assert get_github_repos() == {"joe/myproject"}


def test_get_github_repos_no_github_remotes(fake_git):
    fake_git.add_remote("mygitlab", "git@gitlab.com:joe/myproject.git")
    fake_git.add_remote("upstream", "git@gitlab.com:psf/myproject.git")
    assert get_github_repos() == set()


def test_real_get_github_repos():
    # Since we don't know the name of this repo (forks could be anything),
    # we can't be sure what we get, except it should be word/word, and not end
    # with .git
    repos = get_github_repos()
    assert len(repos) >= 1
    repo = repos.pop()
    assert re.fullmatch(r"[\w-]+/[\w-]+", repo)
    assert not repo.endswith(".git")



"""Updating GitHub releases."""

import logging
import re

import click
import jinja2

from .exceptions import ScrivException
from .github import create_release, get_releases, update_release
from .gitinfo import get_github_repos
from .linkcheck import check_markdown_links
from .scriv import Scriv
from .shell import run_simple_command
from .util import Version, scriv_command

logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "--all",
    "all_entries",
    is_flag=True,
    help="Use all of the changelog entries.",
)
@click.option(
    "--check-links",
    is_flag=True,
    help="Check that links are valid (EXPERIMENTAL).",
)
@click.option(
    "--draft",
    is_flag=True,
    help="Publish a draft release.",
)
@click.option(
    "--dry-run",
    is_flag=True,
    help="Don't post to GitHub, just show what would be done.",
)
@click.option(
    "--fail-if-warn",
    is_flag=True,
    help="Fail if a conversion generates warnings.",
)
@click.option(
    "--repo",
    help="The GitHub repo (owner/reponame) to create the release in.",
)
@scriv_command
def github_release(
    *,
    all_entries: bool,
    check_links: bool,
    draft: bool,
    dry_run: bool,
    fail_if_warn: bool,
    repo: str | None = None,
    config_file: str | None = None,
) -> None:
    """
    Create GitHub releases from the changelog.

    Only the most recent changelog entry is used, unless --all is provided.

    """
    repo = resolve_repo(repo)

    tags = set(map(Version, run_simple_command("git tag").split()))
    releases = {Version(k): v for k, v in get_releases(repo).items()}

    scriv = Scriv(config_file=config_file)
    changelog = scriv.changelog()
    changelog.read()

    for title, sections in changelog.entries().items():
        if title is None:
            continue
        version = Version.from_text(title)
        if version is None:
            logger.warning(f"Entry {title!r} has no version, skipping.")
            continue

        if version not in tags:
            logger.warning(
                f"Version {version} has no tag. No release will be made."
            )
            continue

        section_text = "\n\n".join(sections)
        md = changelog.format_tools().convert_to_markdown(
            section_text, name=title, fail_if_warn=fail_if_warn
        )

        release_data = {
            "body": md,
            "name": str(version),
            "tag_name": str(version),
            "draft": draft,
            "prerelease": version.is_prerelease(),
        }

        ghrel_template = jinja2.Template(scriv.config.ghrel_template)
        md = ghrel_template.render(
            body=md,
            version=version,
            title=title,
            release=release_data,
            config=scriv.config,
        )
        release_data["body"] = md

        if check_links:
            check_markdown_links(md)

        if version in releases:
            release = releases[version]
            if release["body"] != md:
                logger.debug(
                    f"Updating release {version}, data = {release_data}"
                )
                if dry_run:
                    logger.info(f"Would update release {version}")
                    logger.debug(f"Old body:\n{release['body']}")
                    logger.debug(f"New body:\n{md}")
                else:
                    update_release(release, release_data)
        else:
            logger.debug(f"Creating release, data = {release_data}")
            if dry_run:
                logger.info(f"Would create release {version}")
                logger.debug(f"Body:\n{md}")
            else:
                create_release(repo, release_data)

        if not all_entries:
            break


def resolve_repo(repo: str | None) -> str:
    """
    Resolve a repo argument to an owner/repo GitHub repo name.
    """
    if repo is None:
        repos = get_github_repos()
        if len(repos) == 0:
            raise ScrivException("Couldn't find a GitHub repo")
        elif len(repos) > 1:
            repo_list = ", ".join(sorted(repos))
            raise ScrivException(
                f"More than one GitHub repo found: {repo_list}"
            )

        repo = repos.pop()

    if not re.fullmatch(r"[^ /]+/[^ /]+", repo):
        raise ScrivException(f"Repo must be owner/reponame: {repo!r}")

    return repo



"""Markdown text knowledge for scriv."""

import re

from .format import FormatTools, SectionDict


class MdTools(FormatTools):
    """Specifics about how to work with Markdown."""

    def parse_text(
        self,
        text,
    ) -> SectionDict:  # noqa: D102 (inherited docstring)
        lines = text.splitlines()

        # If there's an insert marker, start there.
        for lineno, line in enumerate(lines):
            if self.config.start_marker in line:
                lines = lines[lineno + 1 :]
                break

        sections: SectionDict = {}
        in_comment = False
        paragraphs = None
        section_mark = None

        for line in lines:
            line = line.rstrip()
            if in_comment:
                if re.search(r"-->$", line):
                    in_comment = False
            else:
                if re.search(r"^\s*<!--.*-->$", line):
                    # A one-line comment, skip it.
                    continue
                if re.search(r"""^<a id=(['"])[-.\w]+\1></a>$""", line):
                    # An anchor, we don't need those.
                    continue
                if re.search(r"^\s*<!--", line):
                    in_comment = True
                    continue
                if re.search(r"^#+ ", line):
                    if section_mark is None or line.startswith(section_mark):
                        section_title = line.split(maxsplit=1)[1]
                        paragraphs = sections.setdefault(section_title, [])
                        paragraphs.append("")
                        section_mark = line.partition(" ")[0] + " "
                        continue

                if not line:
                    if paragraphs is not None:
                        paragraphs.append("")
                    continue

                if paragraphs is None:
                    paragraphs = sections.setdefault(None, [])
                    paragraphs.append("")

                paragraphs[-1] += line + "\n"

        # Trim out all empty paragraphs.
        sections = {
            section: [par.rstrip() for par in paragraphs if par]
            for section, paragraphs in sections.items()
            if paragraphs
        }
        return sections

    def format_header(
        self,
        text: str,
        anchor: str | None = None,
    ) -> str:  # noqa: D102 (inherited docstring)
        header = "\n"
        if anchor and self.config.md_html_anchors:
            header += f"<a id='{anchor}'></a>\n"
        if self.config.md_setext_chars:
            header += text + "\n"
            header += self.config.md_setext_chars[0] * len(text) + "\n"
        else:
            num = int(self.config.md_header_level)
            header += "#" * num + " " + text + "\n"
        return header

    def format_sections(
        self,
        sections: SectionDict,
    ) -> str:  # noqa: D102 (inherited docstring)
        lines = [""]
        for section, paragraphs in sections.items():
            if section:
                if self.config.md_setext_chars:
                    lines.append(section)
                    lines.append(self.config.md_setext_chars[1] * len(section))
                else:
                    header_level = int(self.config.md_header_level) + 1
                    lines.append("#" * header_level + " " + section)
                lines.append("")
            for paragraph in paragraphs:
                lines.append(paragraph)
                if not self.config.compact_fragments:
                    lines.append("")
            if self.config.compact_fragments and lines[-1] != "":
                lines.append("")

        return "\n".join(lines)

    def convert_to_markdown(
        self, text: str, name: str = "", fail_if_warn: bool = False
    ) -> str:  # noqa: D102 (inherited docstring)
        return text



"""Collecting fragments."""

import logging
import sys

import click

from .gitinfo import git_add, git_config_bool, git_edit, git_rm
from .scriv import Scriv
from .util import Version, scriv_command

logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "--add/--no-add",
    default=None,
    help="'git add' the updated changelog file and removed fragments.",
)
@click.option(
    "--edit/--no-edit",
    default=None,
    help="Open the changelog file in your text editor.",
)
@click.option(
    "--title", default=None, help="The title text to use for this entry."
)
@click.option(
    "--keep", is_flag=True, help="Keep the fragment files that are collected."
)
@click.option(
    "--version", default=None, help="The version name to use for this entry."
)
@scriv_command
def collect(
    *,
    add: bool | None,
    edit: bool | None,
    title: str,
    keep: bool,
    version: str,
    config_file: str | None,
) -> None:
    """
    Collect and combine fragments into the changelog.
    """
    if title is not None and version is not None:
        sys.exit("Can't provide both --title and --version.")

    if add is None:
        add = git_config_bool("scriv.collect.add")
    if edit is None:
        edit = git_config_bool("scriv.collect.edit")

    scriv = Scriv(config_file=config_file)
    logger.info(f"Collecting from {scriv.config.fragment_directory}")
    frags = scriv.fragments_to_combine()
    if not frags:
        logger.info("No changelog fragments to collect")
        sys.exit(2)

    changelog = scriv.changelog()
    changelog.read()

    if title is None:
        version = Version(version or scriv.config.version)
        if version:
            # Check that we haven't used this version before.
            for etitle in changelog.entries().keys():
                if etitle is None:
                    continue
                eversion = Version.from_text(etitle)
                if eversion is None:
                    sys.exit(
                        f"Entry {etitle!r} is not a valid version! "
                        + "If scriv should ignore this heading, add "
                        + "'scriv-end-here' somewhere before it."
                    )
                if eversion == version:
                    sys.exit(
                        f"Entry {etitle!r} already uses "
                        + f"version {str(version)!r}."
                    )
        new_header = changelog.entry_header(version=version)
    else:
        new_header = changelog.format_tools().format_header(title)

    new_text = changelog.entry_text(scriv.combine_fragments(frags))
    changelog.add_entry(new_header, new_text)
    changelog.write()

    if edit:
        git_edit(changelog.path)

    if add:
        git_add(changelog.path)

    if not keep:
        for frag in frags:
            logger.info(f"Deleting fragment file {str(frag.path)!r}")
            if add:
                git_rm(frag.path)
            else:
                frag.path.unlink()



"""Extracting and checking links."""

import concurrent.futures
import logging
from collections.abc import Iterable

import markdown_it
import requests

logger = logging.getLogger(__name__)


def find_links(markdown_text: str) -> Iterable[str]:
    """Find all the URLs in some Markdown text."""

    def walk_tokens(tokens):
        for token in tokens:
            if token.type == "link_open":
                yield token.attrs["href"]
            if token.children:
                yield from walk_tokens(token.children)

    yield from walk_tokens(markdown_it.MarkdownIt().parse(markdown_text))


def check_markdown_links(markdown_text: str) -> None:
    """
    Check if the URLs in `markdown_text` are reachable.

    Returns None. Logs warnings for unreachable links.
    """
    links = set(find_links(markdown_text))
    with concurrent.futures.ThreadPoolExecutor() as executor:
        executor.map(check_one_link, links)


def check_one_link(url: str) -> None:
    """Check if a URL is reachable. Logs a warning if not."""
    try:
        resp = requests.head(url, timeout=60, allow_redirects=True)
    except Exception as exc:  # pylint: disable=broad-exception-caught
        logger.warning(f"Failed check for {url!r}: {exc}")
        return

    if resp.status_code == 200:
        logger.debug(f"OK link: {url!r}")
    else:
        logger.warning(
            f"Failed check for {url!r}: status code {resp.status_code}"
        )



"""Scriv command-line interface."""

import logging

import click
import click_log

from . import __version__
from .collect import collect
from .create import create
from .ghrel import github_release
from .print import print_

click_log.basic_config(logging.getLogger())


@click.group(
    help=f"""\
        Manage changelogs.  https://scriv.readthedocs.io/

        Version {__version__}
    """
)
@click.version_option()
def cli() -> None:  # noqa: D401
    """The main entry point for the scriv command."""


cli.add_command(create)
cli.add_command(collect)
cli.add_command(github_release)
cli.add_command(print_)



"""Tests of scriv/util.py"""

import pytest

from scriv.util import Version, partition_lines


@pytest.mark.parametrize(
    "text, ver",
    [
        ("v1.2.3 -- 2022-04-06", "v1.2.3"),
        ("Oops, fixed on 6/16/2021.", None),
        ("2022-Apr-06: 12.3-alpha0 finally", "12.3-alpha0"),
        ("2.7.19beta1, 2022-04-08", "2.7.19beta1"),
    ],
)
def test_version_from_text(text, ver):
    if ver is not None:
        ver = Version(ver)
    assert Version.from_text(text) == ver


@pytest.mark.parametrize(
    "version",
    [
        "v1.2.3",
        "17.4.1.3",
    ],
)
def test_is_not_prerelease_version(version):
    assert not Version(version).is_prerelease()


@pytest.mark.parametrize(
    "version",
    [
        "v1.2.3a1",
        "17.4.1.3-beta.2",
    ],
)
def test_is_prerelease_version(version):
    assert Version(version).is_prerelease()


VERSION_EQUALITIES = [
    ("v1.2.3a1", "v1.2.3a1", True),
    ("1.2.3a1", "v1.2.3a1", True),
    ("v1.2.3a1", "1.2.3a1", True),
    ("1.2.3a1", "1.2.3a1", True),
    ("1.2", "1.2.0", False),
    ("1.2.3", "1.2.3a1", False),
    ("1.2.3a1", "1.2.3b1", False),
    ("v1.2.3", "1.2.3a1", False),
]


@pytest.mark.parametrize("ver1, ver2, equal", VERSION_EQUALITIES)
def test_version_equality(ver1, ver2, equal):
    assert (Version(ver1) == Version(ver2)) is equal


@pytest.mark.parametrize("ver1, ver2, equal", VERSION_EQUALITIES)
def test_version_hashing(ver1, ver2, equal):
    assert len({Version(ver1), Version(ver2)}) == (1 if equal else 2)


@pytest.mark.parametrize(
    "text, result",
    [
        ("one\ntwo\nthree\n", ("one\ntwo\nthree\n", "", "")),
        ("oXe\ntwo\nthree\n", ("", "oXe\n", "two\nthree\n")),
        ("one\ntXo\nthree\n", ("one\n", "tXo\n", "three\n")),
        ("one\ntwo\ntXree\n", ("one\ntwo\n", "tXree\n", "")),
        ("one\ntXo\ntXree\n", ("one\n", "tXo\n", "tXree\n")),
    ],
)
def test_partition_lines(text, result):
    assert partition_lines(text, "X") == result



"""Collecting fragments."""

from __future__ import annotations

import logging
import os
import pathlib
import sys

import click

from .scriv import Scriv
from .util import Version, scriv_command

logger = logging.getLogger(__name__)


@click.command(name="print")
@click.option(
    "--version",
    default=None,
    help="The version of the changelog entry to extract.",
)
@click.option(
    "--output",
    type=click.Path(),
    default=None,
    help="The path to a file to write the output to.",
)
@scriv_command
def print_(
    *,
    version: str | None,
    output: pathlib.Path | None,
    config_file: str | None = None,
) -> None:
    """
    Print collected fragments, or print an entry from the changelog.
    """
    scriv = Scriv(config_file=config_file)
    changelog = scriv.changelog()
    newline: str = os.linesep

    if version is None:
        logger.info(f"Generating entry from {scriv.config.fragment_directory}")
        frags = scriv.fragments_to_combine()
        if not frags:
            logger.info("No changelog fragments to collect")
            sys.exit(2)
        contents = changelog.entry_text(scriv.combine_fragments(frags)).strip()
    else:
        logger.info(f"Extracting entry for {version} from {changelog.path}")
        changelog.read()
        newline = changelog.newline
        target_version = Version(version)
        for etitle, sections in changelog.entries().items():
            eversion = Version.from_text(str(etitle))
            if eversion is None:
                continue
            if eversion == target_version:
                contents = f"{changelog.newline * 2}".join(sections).strip()
                break
        else:
            logger.info(f"Unable to find version {version} in the changelog")
            sys.exit(2)

    if output:
        # Standardize newlines to match either the platform default
        # or to match the existing newlines found in the CHANGELOG.
        contents_raw = newline.join(contents.splitlines()).encode("utf-8")
        with open(output, "wb") as file:
            file.write(contents_raw)
    else:
        # Standardize newlines to just '\n' when writing to STDOUT.
        contents = "\n".join(contents.splitlines())
        print(contents)



"""Tests for scriv/format_rst.py."""

import collections
import re
import textwrap

import pytest

from scriv.config import Config
from scriv.exceptions import ScrivException
from scriv.format_rst import RstTools


@pytest.mark.parametrize(
    "text, parsed",
    [
        # Comments are ignored, and the section headers found.
        pytest.param(
            """\
            .. Comments can be here
            .. and here.
            ..
            .. and here.
            Added
            -----

            - This thing was added.
              And we liked it.

            .. More comments can be here
            ..
            .. And here.

            """,
            {"Added": ["- This thing was added.\n  And we liked it."]},
            id="comments_ignored",
        ),
        # Multiple section headers.
        pytest.param(
            """\
            Added
            -----

            - This thing was added.
              And we liked it.


            Fixed
            -----

            - This thing was fixed.

            - Another thing was fixed.

            Added
            -----

            - Also added
              this thing
              that is very important.

            """,
            {
                "Added": [
                    "- This thing was added.\n  And we liked it.",
                    "- Also added\n  this thing\n  that is very important.",
                ],
                "Fixed": [
                    "- This thing was fixed.",
                    "- Another thing was fixed.",
                ],
            },
            id="multiple_headers",
        ),
        # The specific character used for the header line is unimportant.
        pytest.param(
            """\
            Added
            ^^^^^
            - This thing was added.

            Fixed
            ^^^^^
            - This thing was fixed.
            """,
            {
                "Added": ["- This thing was added."],
                "Fixed": ["- This thing was fixed."],
            },
            id="different_underlines",
        ),
        # You can even use periods as the underline, it won't be confused for a
        # comment.
        pytest.param(
            """\
            Fixed
            .....
            - This thing was fixed.

            Added
            .....

            .. a comment.

            - This thing was added.
            """,
            {
                "Added": ["- This thing was added."],
                "Fixed": ["- This thing was fixed."],
            },
            id="period_underline",
        ),
        # It's fine to have no header at all.
        pytest.param(
            """\
            - No header at all.
            """,
            {None: ["- No header at all."]},
            id="no_header",
        ),
        # It's fine to have comments with no header, and multiple bulllets.
        pytest.param(
            """\
            .. This is a scriv fragment.

            - No header at all.

            - Just plain bullets.
            """,
            {None: ["- No header at all.", "- Just plain bullets."]},
            id="no_header_2",
        ),
        # RST syntax is intricate. We understand a subset of it.
        pytest.param(
            """\
            .. _fixed.1:

            Fixed
            .....
            - This thing was fixed: `issue 42`_.

            .. _issue 42: https://github.com/thing/issue/42

            .. _added:

            Added
            .....

            .. a comment.

            - This thing was added.

            .. note::
                This thing doesn't work yet.
                Not sure it ever will... :(

            .. [cite] A citation

            .. |subst| image:: substitution.png

            ..

            """,
            {
                "Added": [
                    "- This thing was added.",
                    (
                        ".. note::\n"
                        + "    This thing doesn't work yet.\n"
                        + "    Not sure it ever will... :("
                    ),
                    ".. [cite] A citation",
                    ".. |subst| image:: substitution.png",
                ],
                "Fixed": [
                    "- This thing was fixed: `issue 42`_.",
                    ".. _issue 42: https://github.com/thing/issue/42",
                ],
            },
            id="intricate_syntax",
        ),
        # A file with only comments and blanks will produce nothing.
        pytest.param(
            """\
            .. Nothing to see here.
            ..

            .. Or here.


            """,
            {},
            id="empty",
        ),
        # Multiple levels of headings only splits on the top-most one.
        pytest.param(
            """\
            =====
            TITLE
            =====

            Irrelevant stuff

            Heading
            =======

            Ignore this

            .. scriv-insert-here

            (prelude)

            ===========
            Section one
            ===========

            subhead
            -------

            In the sub

            subhead 2
            ---------

            Also sub

            Section two
            ===========

            In section two.

            subhead 3
            ---------
            s2s3
            """,
            {
                None: ["(prelude)"],
                "Section one": [
                    "subhead\n-------",
                    "In the sub",
                    "subhead 2\n---------",
                    "Also sub",
                ],
                "Section two": [
                    "In section two.",
                    "subhead 3\n---------\ns2s3",
                ],
            },
            id="multilevel",
        ),
    ],
)
def test_parse_text(text, parsed):
    actual = RstTools().parse_text(textwrap.dedent(text))
    assert actual == parsed


@pytest.mark.parametrize(
    "config_kwargs, sections, expected",
    [
        pytest.param(
            {"rst_header_chars": "#~"},
            [
                (
                    "Added",
                    [
                        "- This thing was added.\n  And we liked it.",
                        "- Also added\n  this thing\n  that is very important.",
                    ],
                ),
                (
                    "Fixed",
                    ["- This thing was fixed.", "- Another thing was fixed."],
                ),
            ],
            """\

            Added
            ~~~~~

            - This thing was added.
              And we liked it.

            - Also added
              this thing
              that is very important.

            Fixed
            ~~~~~

            - This thing was fixed.

            - Another thing was fixed.
            """,
            id="one",
        ),
        pytest.param(
            {"rst_header_chars": "#~"},
            [
                (
                    None,
                    [
                        "- This thing was added.\n  And we liked it.",
                        "- Also added\n  this thing\n  that is very important.",
                    ],
                ),
            ],
            """\

            - This thing was added.
              And we liked it.

            - Also added
              this thing
              that is very important.
            """,
            id="two",
        ),
        pytest.param(
            {"rst_header_chars": "*.", "compact_fragments": True},
            [
                (
                    "Added",
                    [
                        "- This thing was added.\n  And we liked it.",
                        "- Also added\n  this thing\n  that is very important.",
                    ],
                ),
                (
                    "Bugs Fixed",
                    ["- This thing was fixed.", "- Another thing was fixed."],
                ),
            ],
            """\

            Added
            .....

            - This thing was added.
              And we liked it.
            - Also added
              this thing
              that is very important.

            Bugs Fixed
            ..........

            - This thing was fixed.
            - Another thing was fixed.
            """,
            id="compact",
        ),
    ],
)
def test_format_sections(config_kwargs, sections, expected):
    sections = collections.OrderedDict(sections)
    actual = RstTools(Config(**config_kwargs)).format_sections(sections)
    assert actual == textwrap.dedent(expected)


@pytest.mark.parametrize(
    "config_kwargs, text, fh_kwargs, result",
    [
        ({}, "2020-07-26", {}, "\n2020-07-26\n==========\n"),
        (
            {"rst_header_chars": "*-"},
            "2020-07-26",
            {},
            "\n2020-07-26\n**********\n",
        ),
        (
            {},
            "2022-04-03",
            {"anchor": "here"},
            "\n.. _here:\n\n2022-04-03\n==========\n",
        ),
    ],
)
def test_format_header(config_kwargs, text, fh_kwargs, result):
    actual = RstTools(Config(**config_kwargs)).format_header(text, **fh_kwargs)
    assert actual == result


@pytest.mark.parametrize("fail_if_warn", [False, True])
def test_fake_pandoc(fake_run_command, fail_if_warn):
    fake_run_command.patch_module("scriv.format_rst")
    expected_args = [
        "pandoc",
        "-frst",
        "-tmarkdown_strict",
        "--markdown-headings=atx",
        "--wrap=none",
    ]
    if fail_if_warn:
        expected_args.append("--fail-if-warnings")
    expected_text = "The converted text!\nis multi-line\n"

    def fake_pandoc(argv):
        # We got the arguments we expected, plus one more.
        assert argv[:-1] == expected_args
        return (True, expected_text)

    fake_run_command.add_handler("pandoc", fake_pandoc)
    assert (
        RstTools().convert_to_markdown("Hello", fail_if_warn=fail_if_warn)
        == expected_text
    )


def test_fake_pandoc_failing(fake_run_command):
    fake_run_command.patch_module("scriv.format_rst")
    error_text = "There was a problem!!?!"

    def fake_pandoc(argv):  # pylint: disable=unused-argument
        return (False, error_text)

    fake_run_command.add_handler("pandoc", fake_pandoc)
    expected = f"Couldn't convert ReST to Markdown in '':\n{error_text}"
    with pytest.raises(ScrivException, match=re.escape(expected)):
        _ = RstTools().convert_to_markdown("Hello")


@pytest.mark.parametrize(
    "rst_text, md_text",
    [
        (
            """\
            Look at this list:

            - One issue fixed: `issue 123`_.

            - One change merged: `Big change <pull 234_>`_.

            - Improved the `home page <https://example.com/homepage>`_.

            - One more `small change`__.

            .. _issue 123: https://github.com/joe/project/issues/123
            .. _pull 234: https://github.com/joe/project/pull/234
            __ https://github.com/joe/project/issues/999
            """,
            """\
            Look at this list:

            - One issue fixed: [issue 123](https://github.com/joe/project/issues/123).
            - One change merged: [Big change](https://github.com/joe/project/pull/234).
            - Improved the [home page](https://example.com/homepage).
            - One more [small change](https://github.com/joe/project/issues/999).
            """,
        ),
    ],
)
def test_convert_to_markdown(rst_text, md_text):
    converted = RstTools().convert_to_markdown(textwrap.dedent(rst_text))
    # Different versions of pandoc produce slightly different results.  But the
    # markdown is rendered the same regardless of spaces after the
    # bullet-hyphens, so fix them.
    converted = re.sub(r"(?m)^-\s+", "- ", converted)
    expected = textwrap.dedent(md_text)
    assert expected == converted


@pytest.mark.parametrize(
    "rst_text, msg",
    [
        # Various styles of broken links:
        (
            "One issue fixed: `issue 123`_.",
            "Reference not found for 'issue 123'",
        ),
        (
            "One change merged: `Big change <pull 234>_`_.",
            "Reference not found for 'big change <",
        ),
        (
            "Improved the `home page <https://example.com/homepage`_.",
            "Reference not found for 'home page <",
        ),
        # ("One more `small change`__.", "xxx"),     # Hmm, this doesn't error!
        (
            # Not a mistake in RST, but pandoc can't handle it:
            """\
            A big change, thanks to `Jane Contributor <pull
            91_>`_.

            .. _pull 91: https://github.com/joe/project/91
            """,
            "[WARNING] Reference not found for 'pull91'",
        ),
    ],
)
def test_bad_convert_to_markdown(rst_text, msg):
    with pytest.raises(ScrivException, match=re.escape(msg)):
        converted = RstTools().convert_to_markdown(
            textwrap.dedent(rst_text), fail_if_warn=True
        )
        # if we don't get the exception, we can debug the test:
        print(converted)  # pragma: only failure



"""Scriv configuration."""

from __future__ import annotations

import configparser
import contextlib
import logging
import pkgutil
import re
from pathlib import Path
from typing import Any

import attr

from .exceptions import ScrivException
from .literals import find_literal
from .optional import tomllib
from .shell import run_shell_command

logger = logging.getLogger(__name__)


def subset_validator(allowed):
    """Create a validator that ensures a list is a subset of allowed values."""

    def validate_subset(_instance, attribute, value):
        """
        Validate that value is a subset of allowed values.
        """
        value = convert_list(value)
        bad_values = set(value) - set(allowed)
        if bad_values:
            raise ValueError(
                f"{attribute.name!r} must be chosen from: {', '.join(allowed)}."
                + f" {', '.join(map(repr, bad_values))} are not allowed."
            )

    return validate_subset


DEFAULT_FORMAT = "rst"
DEFAULT_CHANGELOG = "CHANGELOG.${config:format}"


@attr.s
class _Options:
    """
    All the settable options for Scriv.
    """

    # The directory for fragments waiting to be collected. Also can have
    # templates and settings for scriv.
    fragment_directory = attr.ib(
        type=str,
        default="changelog.d",
        metadata={
            "doc": """\
                The directory for fragments.  This directory must exist, it
                will not be created.
                """,
        },
    )

    # What format for fragments? reStructuredText ("rst") or Markdown ("md").
    format = attr.ib(  # type: ignore[assignment]
        type=str,
        default=None,
        validator=attr.validators.optional(attr.validators.in_(["rst", "md"])),
        metadata={
            "doc": """\
                The format to use for fragments and for the output changelog
                file.  Can be either "rst" or "md".
                """,
            "doc_default": f"""\
                Derived from the changelog file name if provided,
                otherwise "{DEFAULT_FORMAT}".
                """,
        },
    )

    fragment_name_fields = attr.ib(
        type=list,
        default=["created", "author", "branch"],
        validator=attr.validators.optional(  # type: ignore[arg-type]
            subset_validator(["created", "author", "branch"])
        ),
        metadata={
            "doc": """\
                The components to use when making new fragment file names.
                "created" is a timestamp, "author" is the user nick, and
                "branch" is the current Git branch name if it isn't one of the
                main branch names.
                """,
        },
    )

    # The categories for changelog fragments. Can be empty for no
    # categorization.
    categories = attr.ib(
        type=list,
        default=[
            "Removed",
            "Added",
            "Changed",
            "Deprecated",
            "Fixed",
            "Security",
        ],
        metadata={
            "doc": """\
                Categories to use as headings for changelog items.
                See :ref:`categories`.
                """,
        },
    )

    changelog = attr.ib(
        type=str,
        default=None,
        metadata={
            "doc": """\
                The changelog file managed and read by scriv.  The old name
                for this setting is :ref:`output_file <deprecated_config>`.
                """,
            "doc_default": f"``{DEFAULT_CHANGELOG}``",
        },
    )

    start_marker = attr.ib(
        type=str,
        default="scriv-insert-here",
        metadata={
            "doc": """\
                A marker string indicating where in the changelog file new
                entries should be inserted.  The old name for this setting is
                :ref:`insert_marker <deprecated_config>`.
                """,
        },
    )

    end_marker = attr.ib(
        type=str,
        default="scriv-end-here",
        metadata={
            "doc": """\
                A marker string indicating where in the changelog file the
                changelog ends.
                """,
        },
    )

    # The characters to use for header and section underlines in rst files.
    rst_header_chars = attr.ib(
        type=str,
        default="=-",
        validator=attr.validators.matches_re(r"\S\S"),
        metadata={
            "doc": """\
                Two characters: for reStructuredText changelog files, these
                are the two underline characters to use.  The first is for the
                heading for each changelog entry, the second is for the
                category sections within the entry.
                """,
        },
    )

    # What header level to use for markdown changelog entries?
    md_header_level = attr.ib(
        type=str,
        default="1",
        validator=attr.validators.matches_re(r"[123456]"),
        converter=attr.converters.optional(str),
        metadata={
            "doc": """\
                A number: for Markdown changelog files, this is the heading
                level to use for the entry heading.
                """,
        },
    )

    md_setext_chars = attr.ib(
        type=str,
        default="",
        validator=attr.validators.matches_re(r"^$|\S\S"),
        metadata={
            "doc": """\
                Two characters: for Markdown changelog files, these enable
                setext-style underlined headings if set. The first is for the
                heading for each changelog entry, the second is for the category
                sections within the entry. Use "=-" for standard formatting.
                """,
            "doc_default": "(empty)",
        },
    )

    md_html_anchors = attr.ib(
        type=bool,
        default=True,
        converter=attr.converters.to_bool,
        metadata={
            "doc": """\
                If true, Markdown sections will have HTML anchors added
                for precise linking.
                """,
        },
    )

    compact_fragments = attr.ib(
        type=bool,
        default=False,
        converter=attr.converters.to_bool,
        metadata={
            "doc": """\
                If true, fragments will be concatenated without blank lines
                between them, allowing for compact bullet lists.
                """,
        },
    )

    # The name of the template for new fragments.
    new_fragment_template = attr.ib(
        type=str,
        default="file: new_fragment.${config:format}.j2",
        metadata={
            "doc": """\
                The `Jinja`_ template to use for new fragments.
                """,
        },
    )

    # The template for the title of the changelog entry.
    entry_title_template = attr.ib(
        type=str,
        default=(
            "{% if version %}{{ version }} — {% endif %}"
            + "{{ date.strftime('%Y-%m-%d') }}"
        ),
        metadata={
            "doc": """\
                The `Jinja`_ template to use for the entry heading text for
                changelog entries created by ":ref:`cmd_collect`".
                """,
        },
    )

    # The version string to include in the title if wanted.
    version = attr.ib(
        type=str,
        default="",
        metadata={
            "doc": """\
                The string to use as the version number in the next header
                created by ``scriv collect``.  Often, this will be a
                ``literal:`` directive, to get the version from a string in a
                source file.
                """,
            "doc_default": "(empty)",
        },
    )

    # Branches that aren't interesting enough to use in fragment file names.
    main_branches = attr.ib(
        type=list,
        default=["master", "main", "develop"],
        metadata={
            "doc": """\
                The branch names considered uninteresting to use in new
                fragment file names.
                """,
        },
    )

    # Glob for files in the fragments directory that should not be collected.
    skip_fragments = attr.ib(
        type=str,
        default="README.*",
        metadata={
            "doc": """\
                A glob pattern for files in the fragment directory that should
                not be collected.
                """,
        },
    )

    # Template for GitHub releases
    ghrel_template = attr.ib(
        type=str,
        default="{{body}}",
        metadata={
            "doc": """\
                The template to use for GitHub releases created by the
                ``scriv github-release`` command.

                The extracted Markdown text is available as ``{{body}}``.  You
                must include this to use the text from the changelog file.  The
                version is available as ``{{version}}`` and the title of the
                entry is available as ``{{title}}``.

                The data for the release is available in a ``{{release}}``
                object, including ``{{release.prerelease}}``.  It's a boolean,
                true if this is a pre-release version.

                The scriv configuration is available in a ``{{config}}`` object.
                """,
        },
    )

    def post_create(self):
        """
        Reconcile some interdependent settings after creating the object.
        """
        if self.format is None:
            if self.changelog is not None and "${" not in self.changelog:
                self.format = Path(self.changelog).suffix[1:]
            else:
                self.format = DEFAULT_FORMAT
        if self.changelog is None:
            self.changelog = DEFAULT_CHANGELOG


# Map of old config names to new config names.
DEPRECATED_NAMES = [
    ("output_file", "changelog"),
    ("insert_marker", "start_marker"),
]


@contextlib.contextmanager
def validator_exceptions():
    """
    Context manager for attrs operations that validate.

    Attrs >= 22 says ValueError will have a bunch of arguments, and we only want
    to see the first, and raised as ScrivException.

    """
    try:
        yield
    except ValueError as ve:
        raise ScrivException(f"Invalid configuration: {ve.args[0]}") from ve


class Config:
    """
    Configuration for Scriv.

    All the settable options for Scriv, with resolution of values within other
    values.

    """

    def __init__(self, post_create_=True, **kwargs):
        """All values in _Options can be set as keywords."""
        with validator_exceptions():
            self._options = _Options(**kwargs)
        if post_create_:
            self._options.post_create()

    def __getattr__(self, name):
        """Proxy to self._options, and resolve the value."""
        fields = attr.fields_dict(_Options)
        if name not in fields:
            raise AttributeError(f"Scriv configuration has no {name!r} option")
        attrdef = fields[name]
        value = getattr(self._options, name)
        if attrdef.type is list:
            value = convert_list(value)
        elif isinstance(value, str):
            try:
                value = self.resolve_value(value)
            except ScrivException as se:
                raise ScrivException(
                    f"Couldn't read {name!r} setting: {se}"
                ) from se
        setattr(self, name, value)
        return value

    @classmethod
    def read(cls, file_name: str | None = None) -> Config:
        """
        Read the configuration to use.

        Configuration will be read from setup.cfg, tox.ini, or
        changelog.d/scriv.ini.  If setup.cfg or tox.ini defines
        a new fragment_directory, then scriv.ini is read from there.

        The section can be named ``[scriv]`` or ``[tool.scriv]``.

        """
        config = cls(post_create_=False)
        if file_name is not None:
            config.read_one_config(file_name)
        else:
            config.read_one_config("setup.cfg")
            config.read_one_config("tox.ini")
            config.read_one_toml("pyproject.toml")
            config.read_one_config(
                str(Path(config.fragment_directory) / "scriv.ini")
            )
        with validator_exceptions():
            attr.validate(config._options)
        config._options.post_create()
        return config

    def get_set_option(self, scriv_data, config_name, opt_name):
        """
        Set one option from a config file setting.
        """
        try:
            val: Any = scriv_data[config_name]
        except KeyError:
            pass
        else:
            attrdef = attr.fields_dict(_Options)[opt_name]
            if callable(attrdef.converter):
                val = attrdef.converter(val)
            setattr(self._options, opt_name, val)

    def read_one_config(self, configfile: str) -> None:
        """
        Read one configuration file, adding values to `self`.
        """
        logger.debug(f"Looking for config file {configfile}")
        parser = configparser.ConfigParser()
        files_read = parser.read(configfile, encoding="utf-8")
        if not files_read:
            logger.debug(f"{configfile} doesn't exist")
            return
        logger.debug(f"{configfile} was read")

        section_name = next(
            (
                name
                for name in ["scriv", "tool.scriv"]
                if parser.has_section(name)
            ),
            None,
        )
        if section_name:
            scriv_data = parser[section_name]
            for old, new in DEPRECATED_NAMES:
                self.get_set_option(scriv_data, old, new)
            for attrdef in attr.fields(_Options):
                self.get_set_option(scriv_data, attrdef.name, attrdef.name)

    def read_one_toml(self, tomlfile: str) -> None:
        """
        Read one .toml file if it exists, adding values to `self`.
        """
        logger.debug(f"Looking for config file {tomlfile}")
        tomlpath = Path(tomlfile)
        if not tomlpath.exists():
            logger.debug(f"{tomlfile} doesn't exist")
            return

        toml_text = tomlpath.read_text(encoding="utf-8")
        logger.debug(f"{tomlfile} was read")

        if tomllib is None:
            # Toml support isn't installed. Only print an exception if the
            # config file seems to have settings for us.
            has_scriv = re.search(r"(?m)^\[tool\.scriv\]", toml_text)
            if has_scriv:
                msg = (
                    "Can't read {!r} without TOML support. "
                    + "Install with [toml] extra"
                ).format(tomlfile)
                raise ScrivException(msg)
        else:
            # We have toml installed, parse the file and look for our settings.
            data = tomllib.loads(toml_text)
            try:
                scriv_data = data["tool"]["scriv"]
            except KeyError:
                # No settings for us
                return
            for old, new in DEPRECATED_NAMES:
                self.get_set_option(scriv_data, old, new)
            for attrdef in attr.fields(_Options):
                self.get_set_option(scriv_data, attrdef.name, attrdef.name)

    def resolve_value(self, value: str) -> str:
        """
        Interpret prefixes in config files to find the actual value.

        Also, "${config:format}" is replaced with the configured
        format ("rst" or "md").

        Prefixes:
            "file:" read the content from a file.
            "literal:" read a literal string from a file.
            "command:" read the output of a shell command.

        """
        if self._options.format is not None:
            value = value.replace("${config:format}", self._options.format)
        if value.startswith("file:"):
            file_name = value.partition(":")[2].strip()
            value = self.read_file_value(file_name)
        elif value.startswith("literal:"):
            try:
                _, file_name, literal_name = value.split(":", maxsplit=2)
            except ValueError as ve:
                raise ScrivException(f"Missing value name: {value!r}") from ve
            file_name = file_name.strip()
            if not file_name:
                raise ScrivException(f"Missing file name: {value!r}")
            literal_name = literal_name.strip()
            if not literal_name:
                raise ScrivException(f"Missing value name: {value!r}")
            try:
                found = find_literal(file_name, literal_name)
            except Exception as exc:
                raise ScrivException(
                    f"Couldn't find literal {value!r}: {exc}"
                ) from exc
            if found is None:
                raise ScrivException(
                    f"Couldn't find literal {literal_name!r} in {file_name}: "
                    + f"{value!r}"
                )
            value = found
        elif value.startswith("command:"):
            cmd = value.partition(":")[2].strip()
            ok, out = run_shell_command(cmd)
            if not ok:
                raise ScrivException(f"Command {cmd!r} failed:\n{out}")
            if out.count("\n") == 1:
                out = out.rstrip("\r\n")
            value = out
        return value

    def read_file_value(self, file_name: str) -> str:
        """
        Find the value of a setting that has been specified as a file name.
        """
        value = None
        possibilities = []
        if not re.match(r"\.\.?[/\\]", file_name):
            possibilities.append(Path(self.fragment_directory) / file_name)
        possibilities.append(Path(".") / file_name)

        for file_path in possibilities:
            if file_path.exists():
                value = file_path.read_text()
                break
        else:
            # No path, and doesn't exist: try it as a built-in.
            try:
                file_bytes = pkgutil.get_data("scriv", "templates/" + file_name)
            except OSError:
                pass
            else:
                assert file_bytes
                value = file_bytes.decode("utf-8")

        if value is None:
            raise ScrivException(f"No such file: {file_name}")

        return value


def convert_list(val: str | list[str]) -> list[str]:
    """
    Convert a string value from a config into a list of strings.

    Elements can be separated by commas or newlines.
    """
    if isinstance(val, str):
        vals = re.split(r"[\n,]", val)
        vals = [v.strip() for v in vals]
        vals = [v for v in vals if v]
        return vals
    else:
        return val



"""
Tests of tests/faker.py

Mostly error paths, since the happy paths are covered by other tests.
"""

import re

import pytest

from scriv import shell


def test_no_such_command(fake_run_command):
    assert shell.run_command("wut") == (
        False,
        "no fake command handler: ['wut']",
    )


def test_no_such_git_command(fake_git):
    assert shell.run_command("git hello") == (
        False,
        "no fake git command: ['git', 'hello']",
    )
    assert shell.run_command("git config --wut") == (
        False,
        "no fake git command: ['git', 'config', '--wut']",
    )


@pytest.mark.parametrize(
    "name",
    [
        "foo.bar",
        "foo.bar12",
        "foo.bar-",
        "foo.bar-bar",
        "foo-foo.bar",
        "section.subsection.bar",
        "section.some/sub_section!ok.bar",
    ],
)
def test_git_set_config_good_names(name, fake_git):
    val = "xyzzy plugh!?"
    fake_git.set_config(name, val)
    result = shell.run_command(f"git config --get {name}")
    assert result == (True, val + "\n")


@pytest.mark.parametrize(
    "name",
    [
        "bar",
        "foo.12bar",
        "foo.bar_bar",
        "foo_foo.bar",
    ],
)
def test_git_set_config_bad_names(name, fake_git):
    with pytest.raises(ValueError, match=re.escape(f"invalid key: {name!r}")):
        fake_git.set_config(name, "hello there")



"""Test print logic."""

from textwrap import dedent

import freezegun
import pytest

CHANGELOG_HEADER = dedent("""\

    1.2 - 2020-02-25
    ================
    """)


FRAG = dedent("""\
    Fixed
    -----

    - Launching missiles no longer targets ourselves.
    """)


@pytest.mark.parametrize("newline", ("\r\n", "\n"))
def test_print_fragment(newline, cli_invoke, changelog_d, temp_dir, capsys):
    fragment = FRAG.replace("\n", newline).encode("utf-8")
    (changelog_d / "20170616_nedbat.rst").write_bytes(fragment)

    with freezegun.freeze_time("2020-02-25T15:18:19"):
        cli_invoke(["print"])

    std = capsys.readouterr()
    assert std.out == FRAG


@pytest.mark.parametrize("newline", ("\r\n", "\n"))
def test_print_fragment_output(
    newline, cli_invoke, changelog_d, temp_dir, capsys
):
    fragment = FRAG.replace("\n", newline).encode("utf-8")
    (changelog_d / "20170616_nedbat.rst").write_bytes(fragment)
    output_file = temp_dir / "output.txt"

    with freezegun.freeze_time("2020-02-25T15:18:19"):
        cli_invoke(["print", "--output", output_file])

    std = capsys.readouterr()
    assert std.out == ""
    assert output_file.read_text().strip() == FRAG.strip()


@pytest.mark.parametrize("newline", ("\r\n", "\n"))
def test_print_changelog(newline, cli_invoke, changelog_d, temp_dir, capsys):
    changelog = (CHANGELOG_HEADER + FRAG).replace("\n", newline).encode("utf-8")
    (temp_dir / "CHANGELOG.rst").write_bytes(changelog)

    with freezegun.freeze_time("2020-02-25T15:18:19"):
        cli_invoke(["print", "--version", "1.2"])

    std = capsys.readouterr()
    assert std.out == FRAG


@pytest.mark.parametrize("newline", ("\r\n", "\n"))
def test_print_changelog_output(
    newline, cli_invoke, changelog_d, temp_dir, capsys
):
    changelog = (CHANGELOG_HEADER + FRAG).replace("\n", newline).encode("utf-8")
    (temp_dir / "CHANGELOG.rst").write_bytes(changelog)
    output_file = temp_dir / "output.txt"

    with freezegun.freeze_time("2020-02-25T15:18:19"):
        cli_invoke(["print", "--version", "1.2", "--output", output_file])

    std = capsys.readouterr()
    assert std.out == ""
    assert output_file.read_bytes().decode() == FRAG.strip().replace(
        "\n", newline
    )


def test_print_no_fragments(cli_invoke):
    result = cli_invoke(["print"], expect_ok=False)

    assert result.exit_code == 2
    assert "No changelog fragments to collect" in result.stderr


def test_print_version_not_in_changelog(cli_invoke, changelog_d, temp_dir):
    (temp_dir / "CHANGELOG.rst").write_bytes(b"BOGUS\n=====\n\n1.0\n===")

    result = cli_invoke(["print", "--version", "123.456"], expect_ok=False)

    assert result.exit_code == 2
    assert "Unable to find version 123.456 in the changelog" in result.stderr



"""Changelog and Fragment definitions for Scriv."""

import datetime
import logging
from pathlib import Path

import attr
import jinja2

from .config import Config
from .format import FormatTools, SectionDict, get_format_tools
from .util import partition_lines

logger = logging.getLogger(__name__)


@attr.s
class Fragment:
    """A changelog fragment."""

    path = attr.ib(type=Path)
    format = attr.ib(type=str, default=None)
    content = attr.ib(type=str, default=None)

    def __attrs_post_init__(
        self,
    ):  # noqa: D105 (Missing docstring in magic method)
        if self.format is None:
            self.format = self.path.suffix.lstrip(".")

    def write(self) -> None:
        """Write the content to the file."""
        self.path.write_text(self.content)

    def read(self) -> None:
        """Read the content of the fragment."""
        self.content = self.path.read_text()


@attr.s
class Changelog:
    """A changelog file."""

    path = attr.ib(type=Path)
    config = attr.ib(type=Config)
    newline = attr.ib(type=str, default="")
    text_before = attr.ib(type=str, default="")
    changelog = attr.ib(type=str, default="")
    text_after = attr.ib(type=str, default="")

    def read(self) -> None:
        """Read the changelog if it exists."""
        logger.info(f"Reading changelog {self.path}")
        if self.path.exists():
            with self.path.open("r", encoding="utf-8") as f:
                changelog_text = f.read()
                if f.newlines:  # .newlines may be None, str, or tuple
                    if isinstance(f.newlines, str):
                        self.newline = f.newlines
                    else:
                        self.newline = f.newlines[0]
            before, marker, after = partition_lines(
                changelog_text, self.config.start_marker
            )
            if marker:
                self.text_before = before + marker
                rest = after
            else:
                self.text_before = ""
                rest = before
            self.changelog, marker, after = partition_lines(
                rest, self.config.end_marker
            )
            self.text_after = marker + after
        else:
            logger.warning(f"Changelog {self.path} doesn't exist")

    def format_tools(self) -> FormatTools:
        """Get the appropriate FormatTools for this changelog."""
        return get_format_tools(self.config.format, self.config)

    def entry_header(self, version, date=None) -> str:
        """Format the header for a new entry."""
        title_data = {
            "date": date or datetime.datetime.now(),
            "version": version,
        }
        title_template = jinja2.Template(self.config.entry_title_template)
        new_title = title_template.render(config=self.config, **title_data)
        if new_title.strip():
            anchor = f"changelog-{version}" if version else None
            new_header = self.format_tools().format_header(
                new_title, anchor=anchor
            )
        else:
            new_header = ""
        return new_header

    def entry_text(self, sections: SectionDict) -> str:
        """Format the text of a new entry."""
        return self.format_tools().format_sections(sections)

    def add_entry(self, header: str, text: str) -> None:
        """Add a new entry to the top of the changelog."""
        self.changelog = header + text + self.changelog

    def write(self) -> None:
        """Write the changelog."""
        f = self.path.open("w", encoding="utf-8", newline=self.newline or None)
        with f:
            f.write(self.text_before)
            f.write(self.changelog)
            f.write(self.text_after)

    def entries(self) -> SectionDict:
        """Parse the changelog into a SectionDict."""
        return self.format_tools().parse_text(self.changelog)
